local M = {}

local lpack = require("lpack")

local enableDebug = false

local vehicleData = {}

local function v3_distance(a, b)
    return math.sqrt(math.abs(a.x - b.x)^2 + math.abs(a.y - b.y)^2 + math.abs(a.z - b.z)^2)
end

local function v3_len(a)
    return math.sqrt(math.abs(a.x)^2 + math.abs(a.y)^2 + math.abs(a.z)^2)
end

local function v3_normalize(a)
    local l = v3_len(a)
    local r = {}
    r.x = a.x / l
    r.y = a.y / l
    r.z = a.z / l
    return r
end

local function v3_neg(a)
    local r = {}
    r.x = -a.x
    r.y = -a.y
    r.z = -a.z
    return r
end

local function v3_add(a, b)
    local r = {}
    r.x = a.x + b.x
    r.y = a.y + b.y
    r.z = a.z + b.z
    return r
end

local function v3_sub(a, b)
    local r = {}
    r.x = a.x - b.x
    r.y = a.y - b.y
    r.z = a.z - b.z
    return r
end

local function v3_mul(a, b)
    local r = {}
    r.x = a.x * b.x
    r.y = a.y * b.y
    r.z = a.z * b.z
    return r
end

local function v3_mul_f(a, b)
    local r = {}
    r.x = a.x * b
    r.y = a.y * b
    r.z = a.z * b
    return r
end

local function v3_div(a, b)
    local r = {}
    r.x = a.x / b.x
    r.y = a.y / b.y
    r.z = a.z / b.z
    return r
end

local function v3_div_f(a, b)
    local r = {}
    r.x = a.x / b
    r.y = a.y / b
    r.z = a.z / b
    return r
end

local function v3_dot(a, b)
    return a.x * b.x + a.y * b.y + a.z * b.z
end

local function drawPointCmd(rad, p, color)
    if not enableDebug then return "" end
    local color = color or "color(50, 50, 50, 255)"
    return "obj.debugDrawProxy:drawSphere(" .. rad .. ", vec3(" .. tostring(p.x) .. ", " .. tostring(p.y) .. ", " .. tostring(p.z) .. "), " .. color .. ");"
end

local function drawLineCmd(a, b, color)
    if not enableDebug then return "" end
    local color = color or "color(50, 50, 50, 255)"
    local cmd = "local a = vec3(" .. a.x .. ", " .. a.y .. ", " .. a.z .. ");"
    cmd = cmd .. "local b = vec3(" .. b.x .. ", " .. b.y .. ", " .. b.z .. ");"
    cmd = cmd .. "for i=1,20 do local t = i / 20; local p = vec3(a.x+(b.x-a.x)*t,a.y+(b.y-a.y)*t,a.z+(b.z-a.z)*t); obj.debugDrawProxy:drawSphere(0.1, p, " .. color .. ") end;"
    return cmd
end

local function onUpdate(dtSim, dtRaw)
    if be:getPlayerVehicle(0) == nil then return end
    local cmd = ""
    local playerVehID = be:getPlayerVehicle(0):getID()
    for vehID, data in pairs(vehicleData) do
        cmd = cmd .. drawPointCmd(data.size.x * 0.5, data.position, "color(255, 50, 50, 255)")
    end
    local data = vehicleData[playerVehID]
    if data ~= nil then
        local totalWindStrength = { x = 0, y = 0, z = 0 }
        -- local cmd = "LuuksDraftingMod.drawSphereWrapper(" .. tostring(data.size.x * 0.5) .. ", vec3(" .. tostring(data.position.x) .. ", " .. tostring(data.position.y) .. ", " .. tostring(data.position.z) .. "), color(50, 50, 50, 255))"
        -- be:getPlayerVehicle(0):queueLuaCommand(cmd)
        for vehID, vehData in pairs(vehicleData) do
            if vehID ~= playerVehID and vehData ~= nil then
                local speedCoeff = 3 + (v3_len(vehData.velocity) / 10)
                local dist = v3_distance(data.position, vehData.position)
                local diff = v3_sub(vehData.position, data.position)
                local dot = v3_dot(vehData.vectors.forward, v3_normalize(diff))
                if dot < 0 then dot = 0 end -- Remove negative draft strength possibility
                dot = dot^speedCoeff -- Custom draft strength curve to give a quicker drop off in draft strength when not perfectly in line with the car ahead
                if dot > 1 then dot = 1 end -- Clamp just in case
                dot = dot * v3_dot(vehData.vectors.forward, data.vectors.forward) -- Make sure that if we are facing an entirely different direction, we aren't still getting massive draft strength

                -- Distance coefficient
                local velCoeff = math.max(v3_len(vehData.velocity) / 30, 1)
                local maxDist = 35 * velCoeff
                local distCoeff = 1 - (math.min(dist, maxDist) / maxDist)
                if distCoeff > 1 then distCoeff = 1 end
                if distCoeff < 0 then distCoeff = 0 end
                distCoeff = 1 - distCoeff
                distCoeff = distCoeff * distCoeff
                distCoeff = 1 - distCoeff

                -- Final strength calculation
                local draftStrength = v3_mul_f(v3_normalize(diff), dot)
                draftStrength = v3_mul_f(draftStrength, distCoeff)

                cmd = cmd .. drawLineCmd(data.position, v3_add(data.position, v3_mul_f(draftStrength, 5)))

                local frontalArea = data.size.x * data.size.z
                local aeroForces = v3_mul_f(vehData.velocity, frontalArea)
                local aeroForcesCoeff = v3_dot(data.vectors.forward, v3_normalize(aeroForces)) * v3_len(aeroForces)
                aeroForcesCoeff = aeroForcesCoeff / 20
                local wind = draftStrength
                wind = v3_mul_f(wind, aeroForcesCoeff)

                cmd = cmd .. drawLineCmd(data.position, v3_add(data.position, v3_mul_f(wind, 5)), "color(255,255,50,255)")

                -- cmd = cmd .. "obj:setWind(" .. wind.x .. ", " .. wind.y .. ", " .. wind.z .. ");"
                if not (wind.x == nil or wind.y == nil or wind.z == nil) then
                    totalWindStrength = v3_add(totalWindStrength, wind);
                end
            end
        end

        cmd = cmd .. drawLineCmd(data.position, v3_add(data.position, v3_mul_f(totalWindStrength, 5)), "color(50,255,50,255)")
        cmd = cmd .. "obj:setWind(" .. totalWindStrength.x .. ", " .. totalWindStrength.y .. ", " .. totalWindStrength.z .. ");"

        be:getPlayerVehicle(0):queueLuaCommand(cmd)
    end

    vehicleData = {}
end

local function onVehicleData(dataPacked)
    local data = lpack.decode(dataPacked)
    vehicleData[data.vehicleID] = {}
    -- Copy values over
    for k, v in pairs(data) do
        vehicleData[data.vehicleID][k] = v
    end
end

function onInit()
    print("--------------------------Luuks drafting mod GE loaded!--------------------------")
end

local function toggleDraftDebug()
    enableDebug = not enableDebug
end

M.onInit = onInit
M.onUpdate = onUpdate
M.onVehicleData = onVehicleData
M.toggleDraftDebug = toggleDraftDebug

return M
